==> kelvin_sdk_app-7.7.0-py3-none-any.whl: kelvin/core/application/application.py <==
"""Application interface."""

from __future__ import annotations

from pathlib import Path
from time import time
from typing import Any, Dict, List, Mapping, Optional, Sequence, Type, TypeVar, Union, cast

import yaml

from kelvin.sdk.datatype import Message
from kelvin.sdk.datatype.model import timestamper

from ...sdk.app.mapping_proxy import MappingProxy
from ...sdk.app.utils import build_messages, get_io, inflate
from ..context import ContextInterface

T = TypeVar("T", bound="ApplicationInterface")


class ApplicationInterface:
    """
    The Application Interface is the primary base class for all python
    applications and contains the methods that are publicly available to all
    application types.

    The application allows developers to create, delete, and use timers
    with arbitrary callback methods, emit/publish messages to the
    software bus, and access the current process time.

    """

    def __init__(self, context: ContextInterface) -> None:
        """Initialise application interface."""

        self._context = context

    def on_initialize(
        self,
        configuration: Mapping[str, Any],
        app_configuration: Mapping[str, Any],
        parameters: Optional[Union[Sequence[Any], Mapping[str, Any]]] = None,
    ) -> bool:
        """
        The on_initialize() method is called once at the initial creation of
        the class.

        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        configuration : dict
            The configuration information for the application being executed.
        app_configuration : dict
            The complete application configuration for the application being executed.
        parameters : dict
            The optional parameters information for the application being executed.

        """

        return True  # pragma: no cover

    def on_parameter(self, parameters: Mapping[str, Any]) -> bool:
        """
        The on_parameter() method is called every time the application's
        configuration is modified. This callback will be triggered once after
        initialize() and then once every time the configuration is modified.
        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        parameters : dict
            The parameter information for the application being executed.

        """

        return True  # pragma: no cover

    def _on_parameter(self, assets: Union[Sequence[Any], Mapping[str, Any]]) -> bool:
        """
        The on_parameter() method is called every time the application's
        configuration is modified. This callback will be triggered once after
        initialize() and then once every time the configuration is modified.
        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        parameter : dict
            The parameter information for the application being executed.

        """

        return True  # pragma: no cover

    def _init_parameters(self, parameters: Union[Sequence[Any], Mapping[str, Any]]) -> bool:
        """
        The on_parameter() method is called every time the application's
        configuration is modified. This callback will be triggered once after
        initialize() and then once every time the configuration is modified.
        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        parameters : dict
            The parameter information for the application being executed.

        """

        return True  # pragma: no cover

    def on_terminate(self) -> bool:
        """
        The on_terminate() method is called when the application is being
        terminated.

        This allows application to clean up resources that might have
        been allocated internally, cleanly close out logs, etc. to
        initialize the particular application.

        """

        return True  # pragma: no cover

    def get_process_time(self) -> float:
        """
        Returns the current time of the application.  This time should be used
        by applications for timestamping of messages.  This time will be the
        real wall time by default and the replay time when running in
        simulation mode.

        Returns
        -------
        float : The current process time in seconds.

        """

        return self._context.get_process_time()

    def get_real_time(self) -> float:
        """
        Returns the actual time of the system clock.  This time should be used
        by applications when the actual wall time is required.  This is
        typically used when timestamping sensor measures and computing
        latencies.

        Returns
        -------

        float : The current process time in seconds.

        """

        return self._context.get_real_time()

    def emit(self, data: Message) -> None:
        """
        Takes the incoming data and publishes the contents to the software bus.

        Parameters
        ----------

        data : :obj:`kelvin.sdk.datatype.Message`
            The data to be published on the software bus.

        """

        self._context.emit(data)

    def select(
        self, name: str, start: float = 0.0, end: float = 0.0, limit: int = 1000
    ) -> List[Message]:
        """
        Get a list of metrics from the application storage.

        Accesses the application storage and returns a list of metrics for the specified metric name. The returned
        metrics will be filtered using the start and end dates specified and the number of desired results will be
        limited by the amount of the specified limit.

        Parameters
        ----------
        name : str
            The name of the metric to be looked up in the application storage.
        start : :obj:`float`, optional
            Extract window start.
        end : :obj:`float`, optional
            Extract window end.
        limit : int, optional
            The maximum number of desired data points.

        Returns
        -------
        List: A list of metrics matching the specified filters.

        """

        return self._context.select(name, (start, end), limit)

    def get_input_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the inputs.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_input_registry_map())

    def get_output_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the outputs.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_output_registry_map())

    def get_configuration_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the configuration.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_configuration_registry_map())

    def get_parameter_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the parameters.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_parameter_registry_map())

    @classmethod
    def core_init(
        cls: Type[T],
        configuration: Union[Path, str, Mapping[str, Any]] = "app.yaml",
        startup_time: Optional[float] = None,
        **kwargs: Any,
    ) -> T:
        """Initialise application from Core configuration for testing."""

        from ...sdk.app.context import Context
        from .data import DataApplication

        if isinstance(configuration, str):
            configuration = Path(configuration)

        if isinstance(configuration, Path):
            configuration = yaml.safe_load(configuration.expanduser().resolve().open("rt"))

        if not isinstance(configuration, Mapping):
            raise ValueError("Configuration must be a mapping")

        core_config = MappingProxy({**configuration}).get("app.kelvin", {})

        inputs, outputs, config, params, assets = get_io(core_config)

        expanded = all(isinstance(v, Mapping) and "values" in v for v in config.values())

        if expanded:
            kelvin_app_config = inflate(
                (item["name"], item["value"])
                for item in config.pop("kelvin.app", {}).get("values", [])
            )
        else:
            kelvin_app_config = config.get("kelvin", {}).pop("app", {})

        environment_config = configuration.get("environment", {})
        info_config = configuration.get("info", {})
        kelvin_info = {
            "name": info_config.get("name"),
            "title": info_config.get("title"),
            "description": info_config.get("description"),
            "version": info_config.get("version"),
            "node_name": environment_config.get("node_name"),
            "workload_name": environment_config.get("workload_name"),
        }

        if startup_time is None:
            startup_time = time()

        with timestamper(lambda: int(cast(float, startup_time) * 1e9)):
            init_inputs = build_messages({k: v for k, v in inputs.items() if v.get("values")})
            init_configuration = cast(
                Dict[str, Any], build_messages(config) if expanded else config
            )
            init_configuration.setdefault("kelvin", {}).update(
                {"app": kelvin_app_config, "info": kelvin_info}
            )

        selector_defaults: Dict[str, List[str]] = core_config.get("global", {})
        default = [selector_defaults] if selector_defaults else []
        input_registry_map = {
            name: {
                "name": name,
                "data_type": item["data_type"],
                "control_change": item.get("control_change", False),
                "selectors": [{**selector_defaults, **x} for x in item.get("sources", default)],
            }
            for name, item in inputs.items()
        }
        output_registry_map = {
            name: {
                "name": name,
                "data_type": item["data_type"],
                "control_change": item.get("control_change", False),
                "selectors": [{**selector_defaults, **x} for x in item.get("targets", default)],
            }
            for name, item in outputs.items()
        }
        configuration_registry_map = (
            {
                name: {
                    "name": name,
                    "data_type": item["data_type"],
                    "selectors": [],
                }
                for name, item in config.items()
            }
            if expanded
            else {}
        )
        parameter_registry_map = {
            name: {
                "name": name,
                "data_type": item["data_type"],
                "selectors": item.get("sources", []),
            }
            for name, item in params.items()
        }

        context = Context(
            epoch=startup_time,
            input_registry_map=input_registry_map,
            output_registry_map=output_registry_map,
            configuration_registry_map=configuration_registry_map,
            parameter_registry_map=parameter_registry_map,
        )

        app = cls(context=context, **kwargs)  # type: ignore
        app._init_parameters(params)
        app.on_initialize(
            configuration=init_configuration,
            app_configuration=configuration,
            parameters=assets,
        )
        if isinstance(app, DataApplication) and init_inputs:
            app.on_data([*init_inputs.values()])

        return app

==> kelvin_sdk_app-7.9.0-py3-none-any.whl: kelvin/core/application/application.py <==
"""Application interface."""

from __future__ import annotations

from pathlib import Path
from time import time
from typing import Any, Dict, List, Mapping, Optional, Sequence, Type, TypeVar, Union, cast

import yaml

from kelvin.sdk.datatype import Message
from kelvin.sdk.datatype.model import timestamper

from ...sdk.app.mapping_proxy import MappingProxy
from ...sdk.app.utils import build_messages, get_io, inflate
from ..context import ContextInterface

T = TypeVar("T", bound="ApplicationInterface")


class ApplicationInterface:
    """
    The Application Interface is the primary base class for all python
    applications and contains the methods that are publicly available to all
    application types.

    The application allows developers to create, delete, and use timers
    with arbitrary callback methods, emit/publish messages to the
    software bus, and access the current process time.

    """

    def __init__(self, context: ContextInterface) -> None:
        """Initialise application interface."""

        self._context = context

    def on_initialize(
        self,
        configuration: Mapping[str, Any],
        app_configuration: Mapping[str, Any],
        parameters: Optional[Union[Sequence[Any], Mapping[str, Any]]] = None,
    ) -> bool:
        """
        The on_initialize() method is called once at the initial creation of
        the class.

        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        configuration : dict
            The configuration information for the application being executed.
        app_configuration : dict
            The complete application configuration for the application being executed.
        parameters : dict
            The optional parameters information for the application being executed.

        """

        return True  # pragma: no cover

    def on_parameter(self, parameters: Mapping[str, Any]) -> bool:
        """
        The on_parameter() method is called every time the application's
        configuration is modified. This callback will be triggered once after
        initialize() and then once every time the configuration is modified.
        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        parameters : dict
            The parameter information for the application being executed.

        """

        return True  # pragma: no cover

    def _on_parameter(self, assets: Union[Sequence[Any], Mapping[str, Any]]) -> bool:
        """
        The on_parameter() method is called every time the application's
        configuration is modified. This callback will be triggered once after
        initialize() and then once every time the configuration is modified.
        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        parameter : dict
            The parameter information for the application being executed.

        """

        return True  # pragma: no cover

    def _init_parameters(self, parameters: Union[Sequence[Any], Mapping[str, Any]]) -> bool:
        """
        The on_parameter() method is called every time the application's
        configuration is modified. This callback will be triggered once after
        initialize() and then once every time the configuration is modified.
        The incoming configuration will typically include the default
        parameters and initial conditions necessary to initialize the
        particular application.

        Parameters
        ----------
        parameters : dict
            The parameter information for the application being executed.

        """

        return True  # pragma: no cover

    def on_terminate(self) -> bool:
        """
        The on_terminate() method is called when the application is being
        terminated.

        This allows application to clean up resources that might have
        been allocated internally, cleanly close out logs, etc. to
        initialize the particular application.

        """

        return True  # pragma: no cover

    def get_process_time(self) -> float:
        """
        Returns the current time of the application.  This time should be used
        by applications for timestamping of messages.  This time will be the
        real wall time by default and the replay time when running in
        simulation mode.

        Returns
        -------
        float : The current process time in seconds.

        """

        return self._context.get_process_time()

    def get_real_time(self) -> float:
        """
        Returns the actual time of the system clock.  This time should be used
        by applications when the actual wall time is required.  This is
        typically used when timestamping sensor measures and computing
        latencies.

        Returns
        -------

        float : The current process time in seconds.

        """

        return self._context.get_real_time()

    def emit(self, data: Message) -> None:
        """
        Takes the incoming data and publishes the contents to the software bus.

        Parameters
        ----------

        data : :obj:`kelvin.sdk.datatype.Message`
            The data to be published on the software bus.

        """

        self._context.emit(data)

    def select(
        self, name: str, start: float = 0.0, end: float = 0.0, limit: int = 1000
    ) -> List[Message]:
        """
        Get a list of metrics from the application storage.

        Accesses the application storage and returns a list of metrics for the specified metric name. The returned
        metrics will be filtered using the start and end dates specified and the number of desired results will be
        limited by the amount of the specified limit.

        Parameters
        ----------
        name : str
            The name of the metric to be looked up in the application storage.
        start : :obj:`float`, optional
            Extract window start.
        end : :obj:`float`, optional
            Extract window end.
        limit : int, optional
            The maximum number of desired data points.

        Returns
        -------
        List: A list of metrics matching the specified filters.

        """

        return self._context.select(name, (start, end), limit)

    def get_input_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the inputs.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_input_registry_map())

    def get_output_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the outputs.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_output_registry_map())

    def get_configuration_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the configuration.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_configuration_registry_map())

    def get_parameter_registry_map(self) -> Dict[str, Dict[str, Any]]:
        """
        Get a dict with the registry map of the parameters.

        Returns
        -------
        Dict[str, Dict[str, Any]]: The registry map as a dict

        """

        return yaml.safe_load(self._context.get_parameter_registry_map())

    @classmethod
    def core_init(
        cls: Type[T],
        configuration: Union[Path, str, Mapping[str, Any]] = "app.yaml",
        startup_time: Optional[float] = None,
        **kwargs: Any,
    ) -> T:
        """Initialise application from Core configuration for testing."""

        from ...sdk.app.context import Context
        from .data import DataApplication

        if isinstance(configuration, str):
            configuration = Path(configuration)

        if isinstance(configuration, Path):
            configuration = yaml.safe_load(configuration.expanduser().resolve().open("rt"))

        if not isinstance(configuration, Mapping):
            raise ValueError("Configuration must be a mapping")

        core_config = MappingProxy({**configuration}).get("app.kelvin", {})

        inputs, outputs, config, params, assets = get_io(core_config)

        expanded = all(isinstance(v, Mapping) and "values" in v for v in config.values())

        if expanded:
            kelvin_app_config = inflate(
                (item["name"], item["value"])
                for item in config.pop("kelvin.app", {}).get("values", [])
            )
        else:
            kelvin_app_config = config.get("kelvin", {}).pop("app", {})

        environment_config = configuration.get("environment", {})
        info_config = configuration.get("info", {})
        kelvin_info = {
            "name": info_config.get("name"),
            "title": info_config.get("title"),
            "description": info_config.get("description"),
            "version": info_config.get("version"),
            "node_name": environment_config.get("node_name"),
            "workload_name": environment_config.get("workload_name"),
        }

        if startup_time is None:
            startup_time = time()

        with timestamper(lambda: int(cast(float, startup_time) * 1e9)):
            init_inputs = build_messages({k: v for k, v in inputs.items() if v.get("values")})
            init_configuration = cast(
                Dict[str, Any], build_messages(config) if expanded else config
            )
            init_configuration.setdefault("kelvin", {}).update(
                {"app": kelvin_app_config, "info": kelvin_info}
            )

        default_assets = [asset["name"] for asset in core_config.get("assets", {})]
        selector_defaults: Dict[str, List[str]] = (
            {"asset_names": default_assets} if default_assets else {}
        )
        default = [selector_defaults] if selector_defaults else []
        input_registry_map = {
            name: {
                "name": name,
                "data_type": item["data_type"],
                "control_change": item.get("control_change", False),
                "selectors": [{**selector_defaults, **x} for x in item.get("sources", default)],
            }
            for name, item in inputs.items()
        }
        output_registry_map = {
            name: {
                "name": name,
                "data_type": item["data_type"],
                "control_change": item.get("control_change", False),
                "selectors": [{**selector_defaults, **x} for x in item.get("targets", default)],
            }
            for name, item in outputs.items()
        }
        configuration_registry_map = (
            {
                name: {
                    "name": name,
                    "data_type": item["data_type"],
                    "selectors": [],
                }
                for name, item in config.items()
            }
            if expanded
            else {}
        )
        parameter_registry_map = {
            name: {
                "name": name,
                "data_type": item["data_type"],
                "selectors": item.get("sources", []),
            }
            for name, item in params.items()
        }

        context = Context(
            epoch=startup_time,
            input_registry_map=input_registry_map,
            output_registry_map=output_registry_map,
            configuration_registry_map=configuration_registry_map,
            parameter_registry_map=parameter_registry_map,
        )

        app = cls(context=context, **kwargs)  # type: ignore
        app._init_parameters(params)
        app.on_initialize(
            configuration=init_configuration,
            app_configuration=configuration,
            parameters=assets,
        )
        if isinstance(app, DataApplication) and init_inputs:
            app.on_data([*init_inputs.values()])

        return app
